/* bootpack̃C */

#include "bootpack.h"
#include <stdio.h>

#define HIYOS_VERSION   "HiyOS 0.0.3.20060501"

void make_window8(unsigned char *buf, int xsize, int ysize, char *title);
void putfouts8_asc_sht(struct SHEET *sht, int x, int y, int c, int b, char *s, int l);
void make_textbox8(struct SHEET *sht, int x0, int y0, int sx, int sy, int c);

void io_rdtsc(int *high, int *low);
void io_cpuid(int id_eax, int id_ecx, int *eax, int *ebx, int *ecx, int *edx);

void HariMain(void)
{
	struct BOOTINFO *binfo = (struct BOOTINFO *) ADR_BOOTINFO;
	struct FIFO32 fifo;
	char s[40];
	int fifobuf[128];
	struct TIMER *timer, *timer2, *timer3;
	int mx, my, i, cursor_x, cursor_c;
	unsigned int memtotal;
	struct MOUSE_DEC mdec;
	struct MEMMAN *memman = (struct MEMMAN *) MEMMAN_ADDR;
	struct SHTCTL *shtctl;
	struct SHEET *sht_back, *sht_mouse, *sht_win;
	unsigned char *buf_back, buf_mouse[256], *buf_win;
	static char keytable[0x54] = {
		0,   0,   '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '^', 0,   0,
		'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '@', '[', 0,   0,   'A', 'S',
		'D', 'F', 'G', 'H', 'J', 'K', 'L', ';', ':', 0,   0,   ']', 'Z', 'X', 'C', 'V',
		'B', 'N', 'M', ',', '.', '/', 0,   '*', 0,   ' ', 0,   0,   0,   0,   0,   0,
		0,   0,   0,   0,   0,   0,   0,   '7', '8', '9', '-', '4', '5', '6', '+', '1',
		'2', '3', '0', '.'
	};

/************************/
/* HiyOS 0.0.3.20060501 */
/************************/
	char vendor_string[13];
	char name_string[49];
	char str[256];
	unsigned int eax, ebx, ecx, edx;
	unsigned int family, model, stepping;
	unsigned int family_ex, model_ex, stepping_ex;
	unsigned int max_cpuid, max_cpuid_ex;
	unsigned int tsc_start_high, tsc_start_low;
	unsigned int tsc_end_high, tsc_end_low;
	unsigned int clock;
	
/* HiyOS */

	init_gdtidt();
	init_pic();
	fifo32_init(&fifo, 128, fifobuf);
	init_pit();
	init_keyboard(&fifo, 256);
	enable_mouse(&fifo, 512, &mdec);
	io_out8(PIC0_IMR, 0xf8); /* PITPIC1ƃL[{[h(11111000) */
	io_out8(PIC1_IMR, 0xef); /* }EX(11101111) */

	io_sti(); /* IDT/PIC̏ÎCPŮ荞݋֎~ */

	timer = timer_alloc();
	timer_init(timer, &fifo, 3);
	timer_settime(timer, 300);
	timer2 = timer_alloc();
	timer_init(timer2, &fifo, 2);
	timer_settime(timer2, 200);
	timer3 = timer_alloc();
	timer_init(timer3, &fifo, 1);
	timer_settime(timer3, 50);

	memtotal = memtest(0x00400000, 0xbfffffff);
	memman_init(memman);
	memman_free(memman, 0x00001000, 0x0009e000); /* 0x00001000 - 0x0009efff */
	memman_free(memman, 0x00400000, memtotal - 0x00400000);

	init_palette();
	shtctl = shtctl_init(memman, binfo->vram, binfo->scrnx, binfo->scrny);
	sht_back  = sheet_alloc(shtctl);
	sht_mouse = sheet_alloc(shtctl);
	sht_win   = sheet_alloc(shtctl);
	buf_back  = (unsigned char *) memman_alloc_4k(memman, binfo->scrnx * binfo->scrny);
	buf_win   = (unsigned char *) memman_alloc_4k(memman, 160 * 52);
	sheet_setbuf(sht_back, buf_back, binfo->scrnx, binfo->scrny, -1); /* FȂ */
	sheet_setbuf(sht_mouse, buf_mouse, 16, 16, 99);
	sheet_setbuf(sht_win, buf_win, 160, 52, -1); /* FȂ */
	init_screen8(buf_back, binfo->scrnx, binfo->scrny);
	init_mouse_cursor8(buf_mouse, 99);
	make_window8(buf_win, 160, 52, "window");
	make_textbox8(sht_win, 8, 28, 144, 16, COL8_FFFFFF);
	cursor_x = 8;
	cursor_c = COL8_FFFFFF;
	sheet_slide(sht_back, 0, 0);
	mx = (binfo->scrnx - 16) / 2; /* ʒɂȂ悤ɍWvZ */
	my = (binfo->scrny - 28 - 16) / 2;
	sheet_slide(sht_mouse, mx, my);
	sheet_slide(sht_win, 80, 72);
	sheet_updown(sht_back,  1);
	sheet_updown(sht_win,   0);
	sheet_updown(sht_mouse, 2);
	sprintf(s, "(%3d, %3d)", mx, my);
	putfouts8_asc_sht(sht_back, 0, 0, COL8_FFFFFF, COL8_008484, s, 10);
	sprintf(s, "memory %dMB   free : %dKB",
			memtotal / (1024 * 1024), memman_total(memman) / 1024);
	putfouts8_asc_sht(sht_back, 0, 32, COL8_FFFFFF, COL8_008484, s, 40);

/************************/
/* HiyOS 0.0.3.20060501 */
/************************/

	putfouts8_asc_sht(sht_back, 16, 48, COL8_FFFFFF, COL8_008484,
                      HIYOS_VERSION, 40);
	putfouts8_asc_sht(sht_back, 16, 64, COL8_FFFFFF, COL8_008484,
                      "Project HiyOS based on HariboteOS", 40);

	io_cpuid(0x00000000, 0x0, &eax, &ebx, &ecx, &edx);
	max_cpuid = eax;
	memcpy(vendor_string    , &ebx, 4);
	memcpy(vendor_string + 4, &edx, 4);
	memcpy(vendor_string + 8, &ecx, 4);
	vendor_string[12] = '\0';
	putfouts8_asc_sht(sht_back, 16, 96, COL8_FFFFFF, COL8_008484, vendor_string, 13);

	io_cpuid(0x00000001, 0x0, &eax, &ebx, &ecx, &edx);
	family   = (((eax >> 20) & 0xFF) << 4)
             + ((eax >> 8) & 0xF);
	model    = (((eax >> 16) & 0xF) << 4)
             + ((eax >> 4) & 0xF);
	stepping = (eax >> 0) & 0xF;
	sprintf(str, "Family = %Xh, Model = %Xh, Stepping = %Xh",
            family, model, stepping);
	putfouts8_asc_sht(sht_back, 16, 112, COL8_FFFFFF, COL8_008484, str, 64);

	io_cpuid(0x80000000, 0x0, &eax, &ebx, &ecx, &edx);
	max_cpuid_ex = eax;
	if(max_cpuid_ex >= 0x80000001){
		io_cpuid(0x80000001, 0x0, &eax, &ebx, &ecx, &edx);
		family_ex   = (((eax >> 20) & 0xFF) << 4)
                    + ((eax >> 8) & 0xF);
		model_ex    = (((eax >> 16) & 0xF) << 4)
                    + ((eax >> 4) & 0xF);
		stepping_ex = (eax >> 0) & 0xF;
		sprintf(str, "FamilyEx = %Xh, ModelEx = %Xh, SteppingEx = %Xh",
	            family, model, stepping);
	putfouts8_asc_sht(sht_back, 16, 128, COL8_FFFFFF, COL8_008484, str, 64);

	}
	if(max_cpuid_ex >= 0x80000004){
		io_cpuid(0x80000002, 0x0, &eax, &ebx, &ecx, &edx);
		memcpy(name_string      , &eax, 4);
		memcpy(name_string  +  4, &ebx, 4);
		memcpy(name_string  +  8, &ecx, 4);
		memcpy(name_string  + 12, &edx, 4);
		io_cpuid(0x80000003, 0x0, &eax, &ebx, &ecx, &edx);
		memcpy(name_string  + 16, &eax, 4);
		memcpy(name_string  + 20, &ebx, 4);
		memcpy(name_string  + 24, &ecx, 4);
		memcpy(name_string  + 28, &edx, 4);
		io_cpuid(0x80000004, 0x0, &eax, &ebx, &ecx, &edx);
		memcpy(name_string  + 32, &eax, 4);
		memcpy(name_string  + 36, &ebx, 4);
		memcpy(name_string  + 40, &ecx, 4);
		memcpy(name_string  + 44, &edx, 4);
		name_string[49] = '\0';
		putfouts8_asc_sht(sht_back, 16, 144, COL8_FFFFFF, COL8_008484, name_string, 49);
	}
/* HiyOS */

	for (;;) {
		io_cli();
		if (fifo32_status(&fifo) == 0) {
			io_stihlt();
		} else {
			i = fifo32_get(&fifo);
			io_sti();
			if (256 <= i && i <= 511) { /* L[{[hf[^ */
				sprintf(s, "%02X", i - 256);
				putfouts8_asc_sht(sht_back, 0, 16, COL8_FFFFFF, COL8_008484, s, 2);
				if (i < 0x54 + 256) {
					if (keytable[i - 256] != 0 && cursor_x < 144) { /* ʏ핶 */
						/* ꕶ\ĂAJ[\1i߂ */
						s[0] = keytable[i - 256];
						s[1] = 0;
						putfouts8_asc_sht(sht_win, cursor_x, 28, COL8_000000, COL8_FFFFFF, s, 1);
						cursor_x += 8;
					}
				}
				if (i == 256 + 0x0e && cursor_x > 8) { /* obNXy[X */
					/* J[\Xy[XŏĂAJ[\1߂ */
					putfouts8_asc_sht(sht_win, cursor_x, 28, COL8_000000, COL8_FFFFFF, " ", 1);
					cursor_x -= 8;
				}
				/* J[\̍ĕ\ */
				boxfill8(sht_win->buf, sht_win->bxsize, cursor_c, cursor_x, 28, cursor_x + 7, 43);
				sheet_refresh(sht_win, cursor_x, 28, cursor_x + 8, 44);
			} else if (512 <= i && i <= 767) { /* }EXf[^ */
				if (mouse_decode(&mdec, i - 512) != 0) {
					/* f[^3oCĝŕ\ */
					sprintf(s, "[lcr %4d %4d]", mdec.x, mdec.y);
					if ((mdec.btn & 0x01) != 0) {
						s[1] = 'L';
					}
					if ((mdec.btn & 0x02) != 0) {
						s[3] = 'R';
					}
					if ((mdec.btn & 0x04) != 0) {
						s[2] = 'C';
					}
					putfouts8_asc_sht(sht_back, 32, 16, COL8_FFFFFF, COL8_008484, s, 15);
					/* }EXJ[\̈ړ */
					mx += mdec.x;
					my += mdec.y;
					if (mx < 0) {
						mx = 0;
					}
					if (my < 0) {
						my = 0;
					}
					if (mx > binfo->scrnx - 1) {
						mx = binfo->scrnx - 1;
					}
					if (my > binfo->scrny - 1) {
						my = binfo->scrny - 1;
					}
					sprintf(s, "(%3d, %3d)", mx, my);
					putfouts8_asc_sht(sht_back, 0, 0, COL8_FFFFFF, COL8_008484, s, 10);
					sheet_slide(sht_mouse, mx, my);
					if ((mdec.btn & 0x01) != 0) {
						/* {^ĂAsht_win𓮂 */
						sheet_slide(sht_win, mx - 80, my - 8);
					}
				}
			} else if (i == 3) { /* 3b^C} */
				io_rdtsc(&tsc_end_high, &tsc_end_low);
				clock = (tsc_end_low - tsc_start_low) / 1000 / 1000;
				sprintf(str, "Clock = %dMHz", clock);
				putfouts8_asc_sht(sht_back, 16, 80, COL8_FFFFFF, COL8_008484, str, 24);
			} else if (i == 2) { /* 2b^C} */
			//	putfouts8_asc_sht(sht_back, 0, 80, COL8_FFFFFF, COL8_008484, "3[sec]", 6);
				io_rdtsc(&tsc_start_high, &tsc_start_low);
			} else if (i <= 1) { /* J[\p^C} */
				if (i != 0) {
					timer_init(timer3, &fifo, 0); /* 0 */
					cursor_c = COL8_000000;
				} else {
					timer_init(timer3, &fifo, 1); /* 1 */
					cursor_c = COL8_FFFFFF;
				}
				timer_settime(timer3, 50);
				boxfill8(sht_win->buf, sht_win->bxsize, cursor_c, cursor_x, 28, cursor_x + 7, 43);
				sheet_refresh(sht_win, cursor_x, 28, cursor_x + 8, 44);
			}
		}
	}
}

void make_window8(unsigned char *buf, int xsize, int ysize, char *title)
{
	static char closebtn[14][16] = {
		"OOOOOOOOOOOOOOO@",
		"OQQQQQQQQQQQQQ$@",
		"OQQQQQQQQQQQQQ$@",
		"OQQQ@@QQQQ@@QQ$@",
		"OQQQQ@@QQ@@QQQ$@",
		"OQQQQQ@@@@QQQQ$@",
		"OQQQQQQ@@QQQQQ$@",
		"OQQQQQ@@@@QQQQ$@",
		"OQQQQ@@QQ@@QQQ$@",
		"OQQQ@@QQQQ@@QQ$@",
		"OQQQQQQQQQQQQQ$@",
		"OQQQQQQQQQQQQQ$@",
		"O$$$$$$$$$$$$$$@",
		"@@@@@@@@@@@@@@@@"
	};
	int x, y;
	char c;
	boxfill8(buf, xsize, COL8_C6C6C6, 0,         0,         xsize - 1, 0        );
	boxfill8(buf, xsize, COL8_FFFFFF, 1,         1,         xsize - 2, 1        );
	boxfill8(buf, xsize, COL8_C6C6C6, 0,         0,         0,         ysize - 1);
	boxfill8(buf, xsize, COL8_FFFFFF, 1,         1,         1,         ysize - 2);
	boxfill8(buf, xsize, COL8_848484, xsize - 2, 1,         xsize - 2, ysize - 2);
	boxfill8(buf, xsize, COL8_000000, xsize - 1, 0,         xsize - 1, ysize - 1);
	boxfill8(buf, xsize, COL8_C6C6C6, 2,         2,         xsize - 3, ysize - 3);
	boxfill8(buf, xsize, COL8_000084, 3,         3,         xsize - 4, 20       );
	boxfill8(buf, xsize, COL8_848484, 1,         ysize - 2, xsize - 2, ysize - 2);
	boxfill8(buf, xsize, COL8_000000, 0,         ysize - 1, xsize - 1, ysize - 1);
	putfonts8_asc(buf, xsize, 24, 4, COL8_FFFFFF, title);
	for (y = 0; y < 14; y++) {
		for (x = 0; x < 16; x++) {
			c = closebtn[y][x];
			if (c == '@') {
				c = COL8_000000;
			} else if (c == '$') {
				c = COL8_848484;
			} else if (c == 'Q') {
				c = COL8_C6C6C6;
			} else {
				c = COL8_FFFFFF;
			}
			buf[(5 + y) * xsize + (xsize - 21 + x)] = c;
		}
	}
	return;
}

void putfouts8_asc_sht(struct SHEET *sht, int x, int y, int c, int b, char *s, int l)
{
	boxfill8(sht->buf, sht->bxsize, b, x, y, x + l * 8 - 1, y + 15);
	putfonts8_asc(sht->buf, sht->bxsize, x, y, c, s);
	sheet_refresh(sht, x, y, x + l * 8, y + 16);
	return;
}

void make_textbox8(struct SHEET *sht, int x0, int y0, int sx, int sy, int c)
{
	int x1 = x0 + sx, y1 = y0 + sy;
	boxfill8(sht->buf, sht->bxsize, COL8_848484, x0 - 2, y0 - 3, x1 + 1, y0 - 3);
	boxfill8(sht->buf, sht->bxsize, COL8_848484, x0 - 3, y0 - 3, x0 - 3, y1 + 1);
	boxfill8(sht->buf, sht->bxsize, COL8_FFFFFF, x0 - 3, y1 + 2, x1 + 1, y1 + 2);
	boxfill8(sht->buf, sht->bxsize, COL8_FFFFFF, x1 + 2, y0 - 3, x1 + 2, y1 + 2);
	boxfill8(sht->buf, sht->bxsize, COL8_000000, x0 - 1, y0 - 2, x1 + 0, y0 - 2);
	boxfill8(sht->buf, sht->bxsize, COL8_000000, x0 - 2, y0 - 2, x0 - 2, y1 + 0);
	boxfill8(sht->buf, sht->bxsize, COL8_C6C6C6, x0 - 2, y1 + 1, x1 + 0, y1 + 1);
	boxfill8(sht->buf, sht->bxsize, COL8_C6C6C6, x1 + 1, y0 - 2, x1 + 1, y1 + 1);
	boxfill8(sht->buf, sht->bxsize, c,           x0 - 1, y0 - 1, x1 + 0, y1 + 0);
	return;
}
